/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.core;

import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.createSimpleDescription;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.getTargetConstraintAnnotationAssertionWeaver;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.isAlreadyWoven;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.isAnnotatedByConstraintAnnotation;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.weaveWovenFlag;
import static jp.objectfanatics.commons.javassist.JavassistUtils.createClassPool;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getAnnotations;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getClassPool;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getCtClass;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterTypes;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParametersAnnotations;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getReturnType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getUniqueDeclaredBehaviors;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getUniqueDeclaredMethods;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isVoid;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toBytecode;

import java.io.IOException;

import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.weaver.api.core.Weaver;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalParameterConstraintAnnotationOverridingException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalReturnValueConstraintAnnotationOverridingException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfParameterConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.NoWeavingNeededException;
import jp.objectfanatics.assertion.weaver.api.exception.NotClassFileException;
import jp.objectfanatics.assertion.weaver.impl.constraint.MaxAssertionWeaver;
import jp.objectfanatics.assertion.weaver.impl.constraint.MinAssertionWeaver;
import jp.objectfanatics.assertion.weaver.impl.constraint.NotEmptyAssertionWeaver;
import jp.objectfanatics.assertion.weaver.impl.constraint.NotNullAssertionWeaver;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;

//Checked 09.27.2009. Delete this line before release. とりあえず、次回リリースはこれで行くことにする。
/**
 * Weaver for weaving assertion functions into a class file bytes.
 * 
 * @author Makoto Sato
 */
public class WeaverImpl implements Weaver {
	
	/**
	 * default {@link ConstraintAnnotationAssertionWeaver}s
	 */
	private static final ConstraintAnnotationAssertionWeaver[] defaultConstraintAnnotationAssertionWeavers = {
		new NotNullAssertionWeaver(), new MinAssertionWeaver(), new MaxAssertionWeaver(), new NotEmptyAssertionWeaver()
	};
	
	private final ConstraintAnnotationAssertionWeaver[] constraintAnnotationAssertionWeavers;
	
	/**
	 * Extra classpaths for weaving.
	 */
	private final String[] extraClasspaths;
	
	/**
	 * Initializes an instance.
	 * <p/>
	 * System class paths and <code>extraClasspaths</code> are used as classpaths for weaving.
	 * 
	 * @param constraintAnnotationAssertionWeavers
	 *            {@link ConstraintAnnotationAssertionWeaver}[]
	 * @param extraClasspaths
	 *            extra classpaths
	 * @throws IllegalArgumentException
	 *             if <code>constraintAnnotationAssertionWeavers</code> is <code>null</code>
	 * @throws IllegalArgumentException
	 *             if <code>extraClasspaths</code> is <code>null</code>
	 */
	public WeaverImpl(ConstraintAnnotationAssertionWeaver[] constraintAnnotationAssertionWeavers, String... extraClasspaths) {
		if (constraintAnnotationAssertionWeavers == null) throw new IllegalArgumentException("constraintAnnotationAssertionWeavers is null.");
		if (extraClasspaths == null) throw new IllegalArgumentException("extraClasspaths is null."); // TODO: 未指定の場合に長さ0の配列が来るか要確認。こないなら、nullチェックして空配列を作成しよう。テストケース作成時に対応しませう。
		this.constraintAnnotationAssertionWeavers = constraintAnnotationAssertionWeavers;
		this.extraClasspaths = extraClasspaths;
	}
	
	/**
	 * Initializes an instance.
	 * <p/>
	 * System class paths and <code>extraClasspaths</code> are used as classpaths for weaving.
	 * 
	 * @param extraClasspaths
	 *            extra classpaths
	 * @throws IllegalArgumentException
	 *             if <code>extraClasspaths</code> is <code>null</code>
	 */
	public WeaverImpl(String... extraClasspaths) {
		if (extraClasspaths == null) throw new IllegalArgumentException("extraClasspaths is null."); // TODO: 未指定の場合に長さ0の配列が来るか要確認。こないなら、nullチェックして空配列を作成しよう。テストケース作成時に対応しませう。
		this.extraClasspaths = extraClasspaths;
		this.constraintAnnotationAssertionWeavers = defaultConstraintAnnotationAssertionWeavers;
	}
	
	@Override
	public byte[] weave(byte[] classFileBytes) throws NoWeavingNeededException, NotClassFileException, IllegalUseOfReturnValueConstraintAnnotationException, IllegalUseOfParameterConstraintAnnotationException, IllegalReturnValueConstraintAnnotationOverridingException, IllegalParameterConstraintAnnotationOverridingException {
		if (classFileBytes == null) throw new IllegalArgumentException("classFileBytes is null.");
		try {
			// notice : class pool must not be as an instance field because woven class, it's frozen, can be referred again
			//          and throws frozen exception even if it's just for getting information.
			ClassPool classPool = createClassPool(extraClasspaths);
			
			// get a handle for dealing with the target class file
			CtClass ctClass = getCtClass(classPool, classFileBytes);
			
			// weaves assertions
			return weave(ctClass);
		}
		catch (IOException e) {
			throw new NotClassFileException(e);
		}
	}
	
	/**
	 * Weaves assertion functions if needed.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws NoWeavingNeededException
	 * @throws IllegalUseOfReturnValueConstraintAnnotationException
	 * @throws IllegalUseOfParameterConstraintAnnotationException
	 * @throws IllegalReturnValueConstraintAnnotationOverridingException
	 * @throws IllegalParameterConstraintAnnotationOverridingException
	 */
	private byte[] weave(CtClass ctClass) throws NoWeavingNeededException, IllegalUseOfReturnValueConstraintAnnotationException, IllegalUseOfParameterConstraintAnnotationException, IllegalReturnValueConstraintAnnotationOverridingException, IllegalParameterConstraintAnnotationOverridingException {
		// do nothing if assertion functions are already woven into the class-file-bytes
		if (isAlreadyWoven(ctClass)) throw new NoWeavingNeededException();
		
		boolean isWoven = false;
		isWoven |= weaveAssertionsForParameters(ctClass);
		isWoven |= weaveAssertionsForReturnValue(ctClass);
		
		// weaves @Woven annotation into class file bytes if needed
		if (isWoven) {
			weaveWovenFlag(ctClass);
			return toBytecode(ctClass);
		}
		else {
			throw new NoWeavingNeededException();
		}
	}
	
	/**
	 * Weaves assertion functions for parameters if needed.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws IllegalUseOfParameterConstraintAnnotationException
	 * @throws IllegalParameterConstraintAnnotationOverridingException
	 */
	private boolean weaveAssertionsForParameters(CtClass ctClass) throws IllegalUseOfParameterConstraintAnnotationException, IllegalParameterConstraintAnnotationOverridingException {
		boolean isWoven = false;
		for (CtBehavior ctBehavior : getUniqueDeclaredBehaviors(ctClass)) {
			if (CtMethod.class.isInstance(ctBehavior)) {
				WeaverUtils.assertParameterConstraintAnnotationProperlyOverridden(CtMethod.class.cast(ctBehavior));
			}
			isWoven |= weaveAssertionsForParameters(ctBehavior);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for parameters if needed.
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws IllegalUseOfParameterConstraintAnnotationException
	 */
	private boolean weaveAssertionsForParameters(CtBehavior ctBehavior) throws IllegalUseOfParameterConstraintAnnotationException {
		
		// weave if needed
		boolean isWoven = false;
		for (int paramIndex = 0; paramIndex < getParameterTypes(ctBehavior).length; paramIndex++) {
			isWoven |= weaveAssertionsForAParameter(ctBehavior, paramIndex);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for a parameter if needed.
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @param paramIndex
	 *            target parameter index
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws IllegalUseOfParameterConstraintAnnotationException
	 */
	private boolean weaveAssertionsForAParameter(CtBehavior ctBehavior, int paramIndex) throws IllegalUseOfParameterConstraintAnnotationException {
		
		// temp for clarity
		Annotation[] parameterAnnotations = getParametersAnnotations(ctBehavior)[paramIndex];
		
		// weave assertions for an annotation in a parameter
		boolean isWoven = false;
		for (Annotation parameterAnnotation : parameterAnnotations) {
			isWoven |= weaveAssertionsForAnAnnotationInAParameter(ctBehavior, paramIndex, parameterAnnotation);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for an annotation in a parameter if needed.
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @param paramIndex
	 *            target parameter index
	 * @param parameterAnnotation
	 *            target parameter annotation
	 * @return true if, and only if, assertion functions are woven into the class file bytes
	 * @throws IllegalUseOfParameterConstraintAnnotationException
	 */
	private boolean weaveAssertionsForAnAnnotationInAParameter(CtBehavior ctBehavior, int paramIndex, Annotation parameterAnnotation) throws IllegalUseOfParameterConstraintAnnotationException {
		
		// ignore if annotation is not annotated by ConstraintAnnotation
		ClassPool classPool = ctBehavior.getDeclaringClass().getClassPool();
		if (!isAnnotatedByConstraintAnnotation(classPool, parameterAnnotation)) {
			return false;
		}
		else {
			ConstraintAnnotationAssertionWeaver constraintAnnotationAssertionWeaver = getTargetConstraintAnnotationAssertionWeaver(parameterAnnotation, constraintAnnotationAssertionWeavers);
			constraintAnnotationAssertionWeaver.weaveParameterConstraintAssertion(ctBehavior, paramIndex, parameterAnnotation);
			return true;
		}
	}
	
	/**
	 * Weaves assertion functions for return values if needed.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are woven into the class file bytes
	 * @throws IllegalUseOfReturnValueConstraintAnnotationException
	 * @throws IllegalReturnValueConstraintAnnotationOverridingException
	 */
	private boolean weaveAssertionsForReturnValue(CtClass ctClass) throws IllegalUseOfReturnValueConstraintAnnotationException, IllegalReturnValueConstraintAnnotationOverridingException {
		boolean isWoven = false;
		for (CtMethod ctMethod : getUniqueDeclaredMethods(ctClass)) {
			WeaverUtils.assertReturnValueConstraintAnnotationProperlyOverridden(ctMethod);
			isWoven |= weaveAssertionsForReturnValue(ctMethod);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for return value if needed.
	 * 
	 * @param ctMethod
	 *            a handle for dealing with the target method
	 * @return true if, and only if, assertion functions are woven into the class file bytes
	 * @throws IllegalUseOfReturnValueConstraintAnnotationException
	 */
	private boolean weaveAssertionsForReturnValue(CtMethod ctMethod) throws IllegalUseOfReturnValueConstraintAnnotationException {
		
		// weaves if needed
		boolean isWoven = false;
		Annotation[] annotations = getAnnotations(ctMethod);
		for (Annotation annotation : annotations) {
			isWoven |= weaveAssertionsForAnAnnotationInReturnValue(ctMethod, annotation);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for an annotation in return value if needed.
	 * 
	 * @param ctMethod
	 *            a handle for dealing with the target constructor or method
	 * @param annotation
	 *            target return value constraint annotation
	 * @return true if, and only if, assertion functions are woven into the class file bytes
	 * @throws IllegalUseOfReturnValueConstraintAnnotationException
	 */
	private boolean weaveAssertionsForAnAnnotationInReturnValue(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		
		// TODO: このチェックは、各 ConstraintAnnotationAssertionWeaver ですべき。そうしないと、このメソッドが ConstraintAnnotationAssertionWeaver の事前条件になってしまう。
		
		// ignore if annotation is not annotated by ConstraintAnnotation
		if (!isAnnotatedByConstraintAnnotation(getClassPool(ctMethod), annotation)) {
			return false;
		}
		else {
			ConstraintAnnotationAssertionWeaver constraintAnnotationAssertionWeaver = getTargetConstraintAnnotationAssertionWeaver(annotation, constraintAnnotationAssertionWeavers);
			CtClass returnType = getReturnType(ctMethod);
			if (isVoid(returnType)) {
				// TODO: 文言チェック
				String description = createSimpleDescription(ctMethod, returnType, annotation);
				throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(description, ctMethod, annotation);
			}
			else {
				constraintAnnotationAssertionWeaver.weaveReturnValueConstraintAssertion(ctMethod, annotation);
				return true;
			}
		}
	}
}
