/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.exception;

import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getAnnotationLineBody;
import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getLocationLineValue;
import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getMethodLineBody;
import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getParameterLineBody;
import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getParameterLineTitle;
import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getParameterTypeLineBody;
import static jp.objectfanatics.assertion.weaver.impl.exception.ExceptionUtils.getSourceFileLineValue;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getClassPool;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getSimpleName;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalParameterConstraintAnnotationOverridingException;
import jp.objectfanatics.assertion.weaver.api.exception.info.AnnotationInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.BehaviorInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.MethodInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.ParameterInfo;
import jp.objectfanatics.assertion.weaver.impl.exception.info.AnnotationInfoImpl;
import jp.objectfanatics.assertion.weaver.impl.exception.info.MethodInfoImpl;
import jp.objectfanatics.assertion.weaver.impl.exception.info.ParameterInfoImpl;

//Checked 09.27.2009. Delete this line before release.
/**
 * An implementation class of {@link IllegalParameterConstraintAnnotationOverridingException}.
 * 
 * @author Makoto Sato
 */
public class IllegalParameterConstraintAnnotationOverridingExceptionImpl extends IllegalParameterConstraintAnnotationOverridingException {
	
	private final String description;
	
	private final MethodInfo ancestorMethod;
	
	private final ParameterInfo ancestorParameter;
	
	private final AnnotationInfo ancestorAnnotation;
	
	private final MethodInfo targetMethod;
	
	private final ParameterInfo targetParameter;
	
	/**
	 * Constructor.
	 * 
	 * @param parameterIndex
	 *            parameter index
	 * @param ancestorCtMethod
	 *            ancestor {@link CtMethod}
	 * @param ancestorAnnotation
	 *            target {@link Annotation}
	 * @param targetCtMethod
	 *            target {@link CtMethod}
	 * @throws IllegalArgumentException
	 *             if <code>ancestorCtMethod</code> is <code>null</code>
	 * @throws IllegalArgumentException
	 *             if <code>ancestorAnnotation</code> is <code>null</code>
	 * @throws IllegalArgumentException
	 *             if <code>targetCtMethod</code> is <code>null</code>
	 */
	public IllegalParameterConstraintAnnotationOverridingExceptionImpl(int parameterIndex, CtMethod ancestorCtMethod, Annotation ancestorAnnotation, CtMethod targetCtMethod) {
		if (ancestorCtMethod == null) throw new IllegalArgumentException("ancestorCtMethod is null.");
		if (ancestorAnnotation == null) throw new IllegalArgumentException("targetAnnotation is null.");
		if (targetCtMethod == null) throw new IllegalArgumentException("targetCtMethod is null.");
		// TODO: 今はこのパターンしかありえないので、例外側に持たせておく。将来的にはパターンが増え、再び引数で取得するようになると思われ。s
		this.description = "The following parameter of the overriding method must be annotated by @" + getSimpleName(getClassPool(ancestorCtMethod), ancestorAnnotation) + ".";
		this.ancestorMethod = new MethodInfoImpl(ancestorCtMethod);
		this.ancestorParameter = new ParameterInfoImpl(parameterIndex, ancestorCtMethod);
		this.ancestorAnnotation = new AnnotationInfoImpl(getClassPool(ancestorCtMethod), ancestorAnnotation);
		this.targetMethod = new MethodInfoImpl(targetCtMethod);
		this.targetParameter = new ParameterInfoImpl(parameterIndex, targetCtMethod);
	}
	
	@Override
	public String getDescription() {
		return description;
	}
	
	@Override
	public MethodInfo getAncestorMethod() {
		return ancestorMethod;
	}
	
	@Override
	public ParameterInfo getAncestorParameter() {
		return ancestorParameter;
	}
	
	@Override
	public AnnotationInfo getAncestorAnnotation() {
		return ancestorAnnotation;
	}
	
	@Override
	public MethodInfo getTargetMethod() {
		return targetMethod;
	}
	
	@Override
	public ParameterInfo getTargetParameter() {
		return targetParameter;
	}
	
	@Override
	public BehaviorInfo getTargetBehavior() {
		return targetMethod;
	}
	
	@Override
	public String getDetailedDescription() {
		String ancestorParamTitl = getParameterLineTitle(ancestorParameter, 4, 0);
		String targetParamTitle = getParameterLineTitle(targetParameter, 4, 0);
		StringBuilder sb = new StringBuilder();
		// TODO: リリース直前に、文言の英語的な最終確認。
		sb.append("An expected constraint annotation for a method parameter does not exist " + getLocationLineValue(targetMethod));
		sb.append("  description       : " + description + "\n");
		sb.append("  ancestor method\n");
		sb.append("    location        : " + getLocationLineValue(ancestorMethod));
		sb.append("    source file     : " + getSourceFileLineValue(ancestorMethod));
		sb.append("    method          : " + getMethodLineBody(ancestorMethod));
		sb.append(ancestorParamTitl+ " : " + getParameterLineBody(ancestorParameter));
		sb.append("    parameter type  : " + getParameterTypeLineBody(ancestorParameter));
		sb.append("    annotation      : " + getAnnotationLineBody(ancestorAnnotation));
		sb.append("  overriding method\n");
		sb.append("    location        : " + getLocationLineValue(targetMethod));
		sb.append("    source file     : " + getSourceFileLineValue(targetMethod));
		sb.append("    method          : " + getMethodLineBody(targetMethod));
		sb.append(targetParamTitle + " : " + getParameterLineBody(targetParameter));
		sb.append("    parameter type  : " + getParameterTypeLineBody(targetParameter));
		return sb.toString();
	}
	
	@Override
	public String getMessage() {
		return getDescription();
	}
}
